#lang scheme

(require "structures.scm")
(require "parser.scm")

(provide init-classes! empty-classes
         
         lookup-class lookup-method
         env-get env-set! env-extend env-empty
         field-get field-set! accumulate-fields
         
         union
         )

; internal representation 

(define *classes* '())
(define (empty-classes) (set! *classes* '()))

; class initialization

(define (init-classes! decls)
  (set! *classes* (list 
                   (make-rt-class 'Object #f '() 
                                  ;Object <- this should be the rt-class, not the name
                                  (list (make-rt-method 'Object 'init '()   
                                                        (parse-body '(1)) #f)))
                   ))
  (init-classes-inner! decls))

(define (init-classes-inner! decls)
  (let ((pre-classes (map (lambda (decl)
                            (make-rt-class (class-decl-name decl) 
                                           ; save only the parsing expressions,
                                           ; actual rt values are set in the next step
                                           (class-decl-sname decl) 
                                           (class-decl-fields decl)
                                           (class-decl-methods decl)))
                          decls)))
    (begin
      (set! *classes* (append *classes* pre-classes))
      (map (lambda (class)
             ;(printf "init class: ~a~n" (rt-class-name class))
             (set-rt-class-super!   class (lookup-class (rt-class-super class)))
             (set-rt-class-fields!  class (make-fields  (rt-class-fields class) class))
             (set-rt-class-methods! class (make-methods (rt-class-methods class) class)))
           pre-classes))))

(define (make-fields fdecls class)
  (map (lambda(f) 
         (make-rt-field class 
                        (field-decl-name f)))
       fdecls))

(define (make-methods mdecls class)
  (map (lambda(m) 
         (make-rt-method class 
                         (method-decl-name m)
                         (method-decl-params m)
                         (method-decl-body m)
                         (method-decl-sync m)))
       mdecls))

; environment

(define-struct empty-env ())
(define-struct extended-env (ids values env))

(define (env-get id E)
  (if (empty-env? E) 
      (error 'env-get "Not found ~s." id)
      (let ((idx (list-index id (extended-env-ids E))))
        (if idx
            (begin
              ;(printf "env -> [~a] ~a ... ~a/~a~n" idx id (extended-env-ids E) (extended-env-values E))
              (vector-ref (extended-env-values E) idx))
            (env-get id (extended-env-env E))))))

(define (env-set! id value E)
  (if (empty-env? E) 
      (error 'env-set "Not found ~s." id)
      (let ((idx (list-index id (extended-env-ids E))))
        (if idx
            (vector-set! (extended-env-values E) idx value)
            (env-set! id value (extended-env-env E))))))

(define (env-empty)
  (make-empty-env))

(define (env-extend ids values E)
  (make-extended-env ids (list->vector values) E))


; fields

(define (accumulate-fields class)
  (if (not class)
      '()
      (append (accumulate-fields (rt-class-super class)) 
              (map rt-field-name (rt-class-fields class)))))

(define (field-get obj name class)
  (let ((idx (list-last-index name (accumulate-fields class))))
    (if (not idx)
        (error 'field-get "No such field ~a~n" name)
        (begin
          ;(printf "-> ~a.~a ... ~a/~a~n" obj name value (accumulate-fields class) (rt-object-fvalues obj))
          (vector-ref (rt-object-fvalues obj) idx)))))

(define (field-set! obj name value class)
  (let ((idx (list-last-index name (accumulate-fields class))))
    (if (not idx)
        (error 'field-set "No such field ~a~n" name)
        (begin
          ;(printf "-> ~a.~a = ~a ... ~a/~a~n" obj name value (accumulate-fields class) (rt-object-fvalues obj))
          (vector-set! (rt-object-fvalues obj) idx value)
          value))))

; lookup

(define (lookup-class cname)
  (or (findf (lambda (class) (equal? (rt-class-name class) cname)) *classes*)
      (error 'lookup-class "No class named ~s." cname)))

(define (lookup-method name class)
  (let ((method (lookup-method-no-error name class)))
    (if method
        method
        (error 'lookup-method "No method named ~s starting from class ~a~n." name class))))

(define (lookup-method-no-error name class)
  (if (not class)
      #f
      (let ((method (findf (lambda(m)(equal? (rt-method-name m) name)) (rt-class-methods class))))
        (if method
            method
            (lookup-method name (rt-class-super class))))))

; misc 

(define (list-index v l)
  (if (or (null? l) (not (member v l)))
      #f
      (- (length l) (length (member v l)))))

(define (list-last-index v l)
  (let ((r (list-index v (reverse l))))
    (if (not r)
        r
        (- (- (length l) r) 1))))

(define union
  (lambda (l1 l2)
    (cond 
      ((null? l1) l2)
      ((null? l2) l1)
      ((member (car l1) l2) (union (cdr l1) l2))
      (else (union (cdr l1) (cons (car l1) l2))))))
